#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-11-11
# File  : cisco_meraki_org_wireless_ethernet_statuses.py (check plugin)

# 2024-04-27: made data parsing more robust

from dataclasses import dataclass
from _collections_abc import Mapping

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    Result,
    Service,
    State,
    register,
    render,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
    StringTable,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    get_int,  # type: ignore[import]
    load_json,
)

__ethernet_port_statuses = {
    'aggregation': {
        'enabled': False,
        'speed': 1000
    },
    'name': 'AP01',
    'network': {
        'id': 'L_575897802350012343'
    },
    'ports': [
        {
            'linkNegotiation': {
                'duplex': 'full',
                'speed': 1000
            },
            'name': 'Ethernet 0',
            'poe': {
                'standard': '802.3at'
            }
        }
    ],
    'power': {
        'ac': {
            'isConnected': False
        },
        'mode': 'full',
        'poe': {
            'isConnected': True
        }
    },
    'serial': 'AAAA-BBBB-CCCC'
}


@dataclass(frozen=True)
class WirelessEthernetPortPower:
    mode: str | None
    ac: bool | None
    poe: bool | None


@dataclass(frozen=True)
class WirelessEthernetPort:
    name: str | None
    poe: str | None
    duplex: str | None
    speed: int | None
    power: WirelessEthernetPortPower | None


_is_connected = {
    True: 'Yes',
    False: 'No',
}


def parse_wireless_ethernet_statuses(string_table: StringTable) -> Mapping[str, WirelessEthernetPort] | None:
    json_data = load_json(string_table)
    json_data = json_data[0]

    power = WirelessEthernetPortPower(
        mode=str(json_data['power']['mode']) if json_data['power'].get('mode') is not None else None,
        ac=bool(json_data['power']['ac']['isConnected']) if json_data['power'].get('ac', {}).get(
            'isConnected') is not None else None,
        poe=bool(json_data['power']['poe']['isConnected']) if json_data['power'].get('poe', {}).get(
            'isConnected') is not None else None,
    ) if json_data.get('power') is not None else None

    return {
        port['name']: WirelessEthernetPort(
            name=str(port['name']) if port.get('name') is not None else None,
            power=power,
            duplex=port['linkNegotiation']['duplex'] if port.get('linkNegotiation', {}).get(
                'duplex') is not None else None,
            # changed to bit/s
            speed=int(port['linkNegotiation']['speed']) * 125000 if get_int(port.get('linkNegotiation', {}).get(
                'speed')) else None,
            poe=str(port['poe']['standard']) if port.get('poe', {}).get('standard') is not None else None,
        ) for port in json_data.get('ports', [])
    }


register.agent_section(
    name="cisco_meraki_org_wireless_ethernet_statuses",
    parse_function=parse_wireless_ethernet_statuses,
)


def discover_wireless_ethernet_statuses(section: Mapping[str, WirelessEthernetPort]) -> DiscoveryResult:
    for port in section.keys():
        yield Service(item=port, parameters={'speed': section[port].speed})


def check_wireless_ethernet_statuses(
        item: str,
        params: Mapping[str, any],
        section: Mapping[str, WirelessEthernetPort]
) -> CheckResult:
    def _status_changed(is_state: str, was_state: str, state: int, message: str):
        if is_state != was_state:
            yield Result(state=State(state), notice=f'{message}: is {is_state}, was {was_state}')

    try:
        port: WirelessEthernetPort = section[item]
    except KeyError:
        return None

    if port.speed:
        yield from _status_changed(
            is_state=render.nicspeed(port.speed),
            was_state=render.nicspeed(params['speed']),
            message='Speed changed',
            state=params['state_speed_change']
        )
        yield Result(state=State.OK, summary=f'Speed: {render.nicspeed(port.speed)}')
    else:
        yield Result(state=State(params['state_no_speed']), summary=f'Speed: N/A')

    if port.duplex == 'full':
        yield Result(state=State.OK, summary=f'Duplex: {port.duplex}')
    else:
        yield Result(state=State(params['state_not_full_duplex']), summary=f'Duplex: {port.duplex}')

    if port.power.mode == 'full':
        yield Result(state=State.OK, summary=f'Power mode: {port.power.mode}')
    else:
        yield Result(state=State(params['state_not_on_fill_power']), summary=f'Power mode: {port.power.mode}')

    if port.power.ac:
        yield Result(state=State.OK, summary=f'AC is connected')
    else:
        yield Result(state=State.OK, summary=f'AC is not connected')

    if port.power.poe:
        yield Result(state=State.OK, summary=f'PoE is connected')
    else:
        yield Result(state=State.OK, summary=f'PoE is not connected')

    yield Result(state=State.OK, summary=f'PoE standard: {port.poe}')


register.check_plugin(
    name='cisco_meraki_org_wireless_ethernet_statuses',
    service_name='Port %s',
    discovery_function=discover_wireless_ethernet_statuses,
    check_function=check_wireless_ethernet_statuses,
    check_default_parameters={
        'state_not_on_fill_power': 1,
        'state_not_full_duplex': 1,
        'state_no_speed': 1,
        'state_speed_change': 1,
    },
    check_ruleset_name='cisco_meraki_wireless_ethernet_statuses',
)
