#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-11-11
# File  : cisco_meraki_switch_ports_statuses.py (check plugin)

# 2024-04-08: moved neighbour_name and neighbour_port to key columns
# 2024-04-27: made data parsing more robust
# 2024-05-12: added support for MerakiGetOrganizationSwitchPortsStatusesBySwitch (Early Access)
#             added support for "realtime" traffic counters
#             refactoring parse functions as class method
# 2024-05-20: added discovery rule for port status


from _collections_abc import Mapping, Sequence
from dataclasses import dataclass

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    Result,
    Service,
    State,
    TableRow,
    check_levels,
    register,
    render,
)

from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
    InventoryResult,
    StringTable,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    get_float,  # type: ignore[import]
    get_int,  # type: ignore[import]
    load_json,
)


@dataclass(frozen=True)
class SwitchSecurePort:
    active: bool | None
    authentication_status: str | None
    configOverrides: Mapping[any] | None
    enabled: bool | None

    @classmethod
    def parse(cls, secure_port: Mapping[str, any] | None):
        return cls(
            active=bool(secure_port['active']) if secure_port.get('active') is not None else None,
            authentication_status=str(secure_port['authenticationStatus']) if secure_port.get(
                'authenticationStatus') is not None else None,
            configOverrides=secure_port['configOverrides'] if secure_port.get('configOverrides') is not None else None,
            enabled=bool(secure_port['enabled']) if secure_port.get('enabled') is not None else None,
        ) if secure_port else None


@dataclass(frozen=True)
class SwitchPortCDP:
    address: str | None
    capabilities: str | None
    device_id: str | None
    device_port: str | None
    platform: str | None
    version: str | None

    @classmethod
    def parse(cls, cdp: Mapping[str, str] | None):
        return cls(
            device_id=str(cdp['deviceId']) if cdp.get('deviceId') is not None else None,
            device_port=str(cdp['portId']) if cdp.get('portId') is not None else None,
            address=str(cdp['address']) if cdp.get('address') is not None else None,
            capabilities=str(cdp['capabilities']) if cdp.get('capabilities') is not None else None,
            platform=str(cdp['platform']) if cdp.get('platform') is not None else None,
            version=str(cdp['version']) if cdp.get('version') is not None else None,
        ) if cdp else None


@dataclass(frozen=True)
class SwitchPortLLDP:
    cache_capabilities: str | None
    chassis_id: str | None
    management_address: str | None
    port_description: str | None
    port_id: str | None
    system_description: str | None
    system_name: str | None

    @classmethod
    def parse(cls, lldp: Mapping[str, str] | None):
        return cls(
            chassis_id=str(lldp['chassisId']) if lldp.get('chassisId') is not None else None,
            port_id=str(lldp['portId']) if lldp.get('portId') is not None else None,
            system_name=str(lldp['systemName']) if lldp.get('systemName') is not None else None,
            system_description=str(lldp['systemDescription']) if lldp.get('systemDescription') is not None else None,
            port_description=str(lldp['portDescription']) if lldp.get('portDescription') is not None else None,
            cache_capabilities=str(lldp['systemCapabilities']) if lldp.get('systemCapabilities') is not None else None,
            management_address=str(lldp['managementAddress']) if lldp.get('managementAddress') is not None else None,
        ) if lldp else None


@dataclass(frozen=True)
class SwitchPortUsage:
    total: float
    sent: float
    recv: float

    @classmethod
    def parse(cls, usage: Mapping[str, float] | None):
        """
        Usage in KiloBits -> changed to Bits.
        Args:
            usage: Mapping with the keys 'total', 'sent', 'recv', the values are all int.
                   I.e. {"total": 1944476, "sent": 1099104, "recv": 845372}

        Returns:

        """
        return cls(
            total=get_float(usage.get('total')) * 1000,
            sent=get_float(usage.get('sent')) * 1000,
            recv=get_float(usage.get('recv')) * 1000,
        ) if usage else None


@dataclass(frozen=True)
class SwitchPortTraffic:
    total: float  # The average speed of the data sent and received (in kilobits-per-second).
    sent: float  # The average speed of the data sent (in kilobits-per-second).
    recv: float  # The average speed of the data received (in kilobits-per-second).

    @classmethod
    def parse(cls, traffic: Mapping[str, float] | None):
        """
        A breakdown of the average speed of data that has passed through this port during the timespan.

        input traffic values are  KiloBts/s. Output values are changed to Bits/s
        Args:
            traffic: Mapping with the keys 'total', 'sent', 'recv', the values are all float.
                     I.e. {"total": 184.4, "sent": 104.2, "recv": 80.2}

        Returns:
        """
        return cls(
            total=get_float(traffic.get('total')) * 1000,
            sent=get_float(traffic.get('sent')) * 1000,
            recv=get_float(traffic.get('recv')) * 1000,
        ) if traffic else None


@dataclass(frozen=True)
class SwitchPortSpanningTree:
    status: Sequence[str]

    @classmethod
    def parse(cls, spanning_tree: Mapping[str, Sequence[str]]):
        """
        {"statuses": ["Forwarding", "Is edge", "Is peer-to-peer"]}
        {"statuses": []}
        Args:
            spanning_tree:

        Returns:

        """
        if isinstance(spanning_tree, dict):
            return cls(
                status=[str(status) for status in spanning_tree.get('statuses', [])]
            )


@dataclass(frozen=True)
class SwitchPort:
    cdp: SwitchPortCDP | None
    client_count: int | None
    duplex: str | None
    enabled: bool | None
    errors: Sequence[str] | None
    is_up_link: bool | None
    lldp: SwitchPortLLDP | None
    port_id: str | None
    power_usage_in_wh: float | None
    secure_port: SwitchSecurePort | None
    speed: str | None
    status: str | None
    traffic: SwitchPortTraffic | None
    usage: SwitchPortUsage | None
    warnings: Sequence[str]
    spanning_tree: SwitchPortSpanningTree | None

    @classmethod
    def parse(cls, port: Mapping[str, object]):
        return cls(
            port_id=str(port['portId']) if port.get('portId') is not None else None,
            client_count=get_int(port.get('clientCount')),
            duplex=str(port['duplex']) if port.get('duplex') is not None else None,
            enabled=bool(port['enabled']) if port.get('enabled') is not None else None,
            errors=port['errors'] if port.get('errors') is not None else None,
            is_up_link=bool(port['isUplink']) if port.get('isUplink') is not None else None,
            power_usage_in_wh=get_float(port.get('powerUsageInWh')),
            speed=str(port['speed']) if port.get('speed') is not None else None,
            status=str(port['status']) if port.get('status') is not None else None,
            warnings=port['warnings'] if port.get('warnings') is not None else None,
            lldp=SwitchPortLLDP.parse(port.get('lldp')),
            cdp=SwitchPortCDP.parse(port.get('cdp')),
            secure_port=SwitchSecurePort.parse(port.get('securePort')),
            usage=SwitchPortUsage.parse(port.get('usageInKb')),
            traffic=SwitchPortTraffic.parse((port.get('trafficInKbps'))),
            spanning_tree=SwitchPortSpanningTree.parse(port.get('spanningTree'))
        )


_admin_status = {
    True: 'enabled',
    False: 'disabled',
}

_is_up_link = {
    True: 'yes',
    False: 'no',
}


def parse_switch_ports_statuses(string_table: StringTable) -> Mapping[str, SwitchPort] | None:
    json_data = load_json(string_table)
    json_data = json_data[0]
    # for Early Access MerakiGetOrganizationSwitchPortsStatusesBySwitch
    if isinstance(json_data, dict) and 'ports' in json_data.keys():
        json_data = json_data['ports']

    return {port['portId']: SwitchPort.parse(port) for port in json_data}


register.agent_section(
    name="cisco_meraki_org_switch_ports_statuses",
    parse_function=parse_switch_ports_statuses,
)


def discover_switch_ports_statuses(params: Mapping[str, object], section: Mapping[str, SwitchPort]) -> DiscoveryResult:
    discovered_port_states = params['discovered_port_states']
    for port in section.keys():
        if section[port].enabled in discovered_port_states and section[port].status in discovered_port_states:
            yield Service(
                item=port,
                parameters={
                    'enabled': section[port].enabled,
                    'status': section[port].status,
                    'speed': section[port].speed,
                }
            )


def check_switch_ports_statuses(item: str, params: Mapping[str, any], section: Mapping[str, SwitchPort]) -> CheckResult:
    def _status_changed(is_state: str, was_state: str, state: int, message: str):
        if is_state != was_state:
            is_state = is_state if is_state else 'N/A'
            was_state = was_state if was_state else 'N/A'
            yield Result(state=State(state), notice=f'{message}: from {was_state}, to {is_state}')

    try:
        port: SwitchPort = section[item]
    except KeyError:
        return

    # check admin state changed
    yield from _status_changed(
        is_state=_admin_status[port.enabled],
        was_state=_admin_status[params['enabled']],
        message='Admin status changed',
        state=params['state_admin_change'],
    )

    if port.enabled:  # check admin sate
        yield Result(state=State.OK, notice=f'Admin status: {_admin_status[port.enabled]}')
        # check operational status changed
        yield from _status_changed(
            is_state=port.status.lower(),
            was_state=params['status'].lower(),
            message='Operational status changed',
            state=params['state_op_change']
        )
        if port.status.lower() == 'connected':  # check operational state
            yield Result(state=State.OK, notice=f'Operational status: {port.status}')
            # check speed changed
            yield from _status_changed(
                is_state=port.speed,
                was_state=params['speed'],
                message='Speed changed',
                state=params['state_speed_change']
            )
            if params['speed'] == port.speed:  # only if speed unchanged
                yield Result(state=State.OK, summary=f'Speed: {port.speed}')
            if port.duplex.lower() == 'full':  # check duplex state
                yield Result(state=State.OK, summary=f'Duplex: {port.duplex}')
            else:
                yield Result(state=State(params['state_not_full_duplex']), summary=f'Duplex: {port.duplex}')
            yield Result(state=State.OK, summary=f'Clients: {port.client_count}')

            if params.get('show_traffic'):
                yield from check_levels(
                    value=port.traffic.sent,  # Bits
                    label='Out',
                    metric_name='if_out_bps',
                    render_func=lambda v: render.networkbandwidth(v/8),  # Bytes
                    # notice_only=True,
                )
                yield from check_levels(
                    value=port.traffic.recv,  # Bits
                    label='In',
                    metric_name='if_in_bps',
                    render_func=lambda v: render.networkbandwidth(v/8),  # Bytes
                    # notice_only=True,
                )
        else:
            yield Result(state=State(params['state_not_connected']), summary=f'Operational status: {port.status}')
    else:
        yield Result(state=State(params['state_disabled']), summary=f'Admin status: {_admin_status[port.enabled]}')

    if port.is_up_link:
        yield Result(state=State.OK, summary=f'UP-Link: {_is_up_link[port.is_up_link]}')
    else:
        yield Result(state=State.OK, notice=f'UP-Link: {_is_up_link[port.is_up_link]}')

    if port.power_usage_in_wh:
        yield Result(state=State.OK, summary=f'Power usage: {port.power_usage_in_wh} Wh')

    if port.spanning_tree:
        for status in port.spanning_tree.status:
            yield Result(state=State.OK, notice=f'Spanning-tree status: {status}')

    for warning in port.warnings:
        yield Result(state=State.WARN, notice=f'{warning}')
    for error in port.errors:
        if error not in ['Port disconnected', 'Port disabled']:
            yield Result(state=State.CRIT, notice=f'{error}')

    if port.secure_port and port.secure_port.enabled:
        yield Result(state=State.UNKNOWN, summary=f'Secure Port enabled', details=f'Secure Port: {port.secure_port}')


register.check_plugin(
    name='cisco_meraki_org_switch_ports_statuses',
    service_name='Port %s',
    discovery_function=discover_switch_ports_statuses,
    check_function=check_switch_ports_statuses,
    check_default_parameters={
        'state_disabled': 0,
        'state_not_connected': 0,
        'state_not_full_duplex': 1,
        'state_speed_change': 1,
        'state_admin_change': 1,
        'state_op_change': 1,
    },
    check_ruleset_name='cisco_meraki_switch_ports_statuses',
    discovery_ruleset_name='discovery_cisco_meraki_switch_ports_statuses',
    discovery_default_parameters={
        'discovered_port_states': [True, False, 'Connected', 'Disconnected']
    }
)


def inventory_meraki_cdp_cache(section: Mapping[str, SwitchPort]) -> InventoryResult:
    path = ['networking', 'cdp_cache', 'neighbours']

    for port in section.values():
        if cdp := port.cdp:
            key_columns = {
                'local_port': port.port_id,
                'neighbour_name': cdp.device_id,
                'neighbour_port': cdp.device_port,
            }
            neighbour = {
                'neighbour_address': cdp.address,
                'platform_details': cdp.version,
                'platform': cdp.platform,
                'capabilities': cdp.capabilities
            }
            yield TableRow(
                path=path,
                key_columns=key_columns,
                inventory_columns=neighbour
            )


register.inventory_plugin(
    name='inv_meraki_cdp_cache',
    sections=['cisco_meraki_org_switch_ports_statuses'],
    inventory_function=inventory_meraki_cdp_cache,
)


def inventory_meraki_lldp_cache(section: Mapping[str, SwitchPort]) -> InventoryResult:
    path = ['networking', 'lldp_cache', 'neighbours']

    for port in section.values():
        if lldp := port.lldp:
            key_columns = {
                'local_port': port.port_id,
                'neighbour_name': lldp.system_name,
                'neighbour_port': lldp.port_id,
            }
            neighbour = {
                'neighbour_id': lldp.chassis_id,
                'system_description': lldp.system_description,
                'port_description': lldp.port_description,
                'capabilities': lldp.cache_capabilities,
                'neighbour_address': lldp.management_address,
            }
            yield TableRow(
                path=path,
                key_columns=key_columns,
                inventory_columns=neighbour
            )


register.inventory_plugin(
    name='inv_meraki_lldp_cache',
    sections=['cisco_meraki_org_switch_ports_statuses'],
    inventory_function=inventory_meraki_lldp_cache,
)
