#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-02-02
# File  : cisco_meraki_switch_ports_statuses.py (WATO)

# 2024-05-12: added support for MerakiGetOrganizationSwitchPortsStatusesBySwitch (Early Access)
#             added traffic counters as perfdata
# 2024-05-19: reworked switch port traffic
# 2024-05-20: added discovery rule for port status

from cmk.gui.i18n import _
from cmk.gui.plugins.wato.utils import (
    CheckParameterRulespecWithItem,
    HostRulespec,
    RulespecGroupCheckParametersDiscovery,
    RulespecGroupCheckParametersNetworking,
    rulespec_registry,
)
from cmk.gui.valuespec import (
    Dictionary,
    ListChoice,
    FixedValue,
    MonitoringState,
    TextInput,
)

from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    _SEC_NAME_SWITCH_PORTS_STATUSES,   # type: ignore[import]
    # Early Access
    _SEC_NAME_ORG_SWITCH_PORTS_STATUSES,   # type: ignore[import]
)


def _parameter_valuespec_cisco_meraki_switch_ports_statuses():
    return Dictionary(
        title=_('Cisco Meraki Appliance Uplinks'),
        optional_keys=True,
        elements=[
            ('state_disabled',
             MonitoringState(
                 title=_('Monitoring state if port is "disabled"'),
                 default_value=0,
             )),
            ('state_not_connected',
             MonitoringState(
                 title=_('Monitoring state if port is "not connected"'),
                 default_value=0,
             )),
            ('state_not_full_duplex',
             MonitoringState(
                 title=_('Monitoring state if port is "not full duplex"'),
                 default_value=1,
             )),
            ('state_speed_change',
             MonitoringState(
                 title=_('Monitoring state if speed is changed'),
                 default_value=1,
             )),
            ('state_admin_change',
             MonitoringState(
                 title=_('Monitoring state if admin state is changed'),
                 default_value=1,
             )),
            ('state_op_change',
             MonitoringState(
                 title=_('Monitoring state if operational state is changed'),
                 default_value=1,
             )),
            ('show_traffic',
             FixedValue(
                 True,
                 title=_('Show bandwidth (use only with cache disabled)'),
                 totext='Bandwidth monitoring enabled',
                 help=_(
                     'Use only with cache disabled in the Meraki special agent settings. '
                     'Depending on your Meraki organization size (in terms of number of switches) '
                     'this will exceeds the limits of the allowed API requests per second. You can try to '
                     'enable "Early Access" in the Meraki dashboard. In the Meraki special agent settings '
                     f'switch from "{_SEC_NAME_SWITCH_PORTS_STATUSES}" to "{_SEC_NAME_ORG_SWITCH_PORTS_STATUSES}". '
                     'This will fetch all the switch data with one API request instead of one request for each switch.'
                 ),
             )),
        ],
    )


rulespec_registry.register(
    CheckParameterRulespecWithItem(
        title=lambda: _('Cisco Meraki Switch Ports'),
        check_group_name='cisco_meraki_switch_ports_statuses',
        group=RulespecGroupCheckParametersNetworking,
        parameter_valuespec=_parameter_valuespec_cisco_meraki_switch_ports_statuses,
        match_type='dict',
        item_spec=lambda: TextInput(title=_('The Port ID'), ),
    )
)



def _valuespec_discovery_cisco_meraki_switch_ports_statuses():
    return Dictionary(
        title=_('Cisco Meraki Switch Ports'),
        elements=[
            ('discovered_port_states',
             ListChoice(
                 title=_('Select Ports to discover'),
                 choices=[
                     (True, _('Admin enabled')),
                     (False, _('Admin disabled')),
                     ('Connected', _('Connected')),
                     ('Disconnected', _('Disconnected')),
                 ],
                 help=_('Select the port states for discovery'),
                 default_value=[True, False, 'Connected', 'Disconnected'],
             )),
        ],
        required_keys=['item_variant'],
    )


rulespec_registry.register(
    HostRulespec(
        group=RulespecGroupCheckParametersDiscovery,
        match_type='dict',
        name='discovery_cisco_meraki_switch_ports_statuses',
        valuespec=_valuespec_discovery_cisco_meraki_switch_ports_statuses,
    ))