#!/usr/bin/env python3
# Copyright (C) 2022 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

#
# NOTE: to use the updated cisco meraki special agent WATO via ~/local structure you need to remove/rename the
#       original file (~/lib/check_mk/gui/plugins/wato/special_agents/cisco_meraki.py). At the moment you
#       can not supersede a special agent wato file via ~/local structure :-(
#

# enhancements by thl-cmk[at]outlook[dot]com, https://thl-cmk.hopto.org
# - added check for duplicate organisation IDs
# - added some (basic) online help
# - changed clarified title of "orgs" from "Organisations" to "Organisation IDs"
# - added section titles from utils/cisco_meraki.py (reuse)
# - added option for host prefix/suffix/case per organisation -> needed for dynamic host management
# - added section names from utils/cisco_meraki.py (reuse with special agent)
# 2023-11-18: moved from ~/local/lib/check_mk/gui/plugins/wato to ~/local/share/check_mk/web/plugins/wato
#             changed sections to excluded_sections
# 2023-11-22: replaced host_suffix_prefix option by org_id_as_prefix
#             changed excluded_sections option from DualListChoice to ListChoice to avoid the "Selected" header
#             in conjunction with "excluded Sections" title
# 2024-05-15: added api_key to required_keys
# 2024-06-23: added cache time per section -> not nice but should work.

from typing import List

from cmk.gui.i18n import _
from cmk.gui.plugins.wato.special_agents.common import (
    RulespecGroupDatasourceProgramsApps
)
from cmk.gui.plugins.wato.utils import (
    HostRulespec,
    HTTPProxyReference,
    IndividualOrStoredPassword,
    MKUserError,
    rulespec_registry,
)
from cmk.gui.valuespec import (
    Dictionary,
    FixedValue,
    Integer,
    ListChoice,
    ListOfStrings,
    Tuple,
    ValueSpec,
)
from cmk.base.plugins.agent_based.utils.cisco_meraki import (
    # _SEC_NAME_DEVICE_INFO,
    # _SEC_NAME_NETWORKS,
    # _SEC_NAME_ORGANISATIONS,
    _SEC_NAME_ORG_API_REQUESTS,  # type: ignore[import]
    _SEC_NAME_LICENSES_OVERVIEW,  # type: ignore[import]
    _SEC_NAME_DEVICE_STATUSES,  # type: ignore[import]
    _SEC_NAME_SENSOR_READINGS,  # type: ignore[import]
    _SEC_NAME_DEVICE_UPLINKS_INFO,  # type: ignore[import]
    _SEC_NAME_APPLIANCE_UPLINKS,  # type: ignore[import]
    _SEC_NAME_APPLIANCE_UPLINKS_USAGE,  # type: ignore[import]
    _SEC_NAME_APPLIANCE_VPNS,  # type: ignore[import]
    _SEC_NAME_APPLIANCE_PERFORMANCE,
    _SEC_NAME_SWITCH_PORTS_STATUSES,  # type: ignore[import]
    _SEC_NAME_WIRELESS_ETHERNET_STATUSES,  # type: ignore[import]
    _SEC_NAME_WIRELESS_DEVICE_STATUS,  # type: ignore[import]
    _SEC_NAME_CELLULAR_UPLINKS,  # type: ignore[import]
    # Early Access
    _SEC_NAME_ORG_SWITCH_PORTS_STATUSES,  # type: ignore[import]
)

_SEC_TITLE_DEVICE_INFO = _('Device info (Organization)')
_SEC_TITLE_NETWORKS = _('Network info (Organization)')
_SEC_TITLE_ORGANISATIONS = _('Organization (Agent)')
_SEC_TITLE_ORG_API_REQUESTS = _('API request (Organizaion)')
_SEC_TITLE_APPLIANCE_UPLINKS = _('Appliances uplinks (Organizaion)')
_SEC_TITLE_APPLIANCE_UPLINKS_USAGE = _(
    'Appliances uplinks usage (Organizaion)')
_SEC_TITLE_APPLIANCE_VPNS = _('Appliances VPNs (Organizaion)')
_SEC_TITLE_APPLIANCE_PERFORMANCE = _('Appliances Utilization (Device)')
_SEC_TITLE_CELLULAR_UPLINKS = _('Cellular devices uplinks (Organizaion)')
_SEC_TITLE_DEVICE_STATUSES = _('Devices status (Organizaion)')
_SEC_TITLE_DEVICE_UPLINKS_INFO = _('Devices uplink info (Organizaion)')
_SEC_TITLE_LICENSES_OVERVIEW = _('Licenses overview (Organizaion)')
_SEC_TITLE_SENSOR_READINGS = _('Sensors readings (Organizaion)')
_SEC_TITLE_SWITCH_PORTS_STATUSES = _('Switch ports status (Device)')
_SEC_TITLE_WIRELESS_ETHERNET_STATUSES = _(
    'Wireless devices ethernet status (Organizaion)')
_SEC_TITLE_WIRELESS_DEVICE_STATUS = _('Wireless devices SSIDs status (Device)')
_SEC_TITLE_ORG_SWITCH_PORTS_STATUSES = _(
    'Switch port status (Organizaion/Early Access)')


def _validate_orgs(value: List[str] | None, var_prefix: str):
    # Check for duplicate Organisations
    if value is None:
        return
    _p = list(set(value.copy()))
    if len(_p) != len(value):
        raise MKUserError(var_prefix, _('Duplicate Organisation found'))

    for org_id in value:
        if not org_id.isdigit():
            raise MKUserError(
                var_prefix, _(
                    f'Not a valid Organisation ID {org_id}. Organisation IDs'
                    ' are all digits'
                )
            )


def _valuespec_special_agent_cisco_meraki() -> ValueSpec:
    return Dictionary(
        title=_('Cisco Meraki'),
        elements=[
            ('api_key', IndividualOrStoredPassword(
                title=_('API Key'),
                allow_empty=False,
                help=_('The key to access the Cisco Meraki Cloud Rest API.')
            )),
            ('proxy', HTTPProxyReference(),),
            ('no_cache', FixedValue(
                value=True,
                title=_('Disable Cache'),
                totext=_(''),
                help=_(
                    'Never use cached information. By default the agent will cache received '
                    'data to avoid API limits and speed up the data retrievel.'
                )
            )),
            ('org_id_as_prefix', FixedValue(
                value=True,
                title=_('Uese organisation ID as host prefix'),
                totext=_(''),
                help=_(
                    'The organisation ID will be used as prefix for the hostname (separated by a "\'"). Use '
                    'this option together with a "Hostname translation for piggybacked hosts" to add a organisation '
                    'prefix to the hosts from the Cisco Meraki cloud to avoid conflicting hostnames. You can also use '
                    'this option along with the "Dynamic host management" to sort the host in organisation specific '
                    'folders.'
                )
            )),
            ('excluded_sections',
             ListChoice(
                 title=_('excluded Sections'),
                 choices=[
                     (_SEC_NAME_ORG_API_REQUESTS, _SEC_TITLE_ORG_API_REQUESTS),
                     (_SEC_NAME_APPLIANCE_UPLINKS, _SEC_TITLE_APPLIANCE_UPLINKS),
                     (_SEC_NAME_APPLIANCE_UPLINKS_USAGE,
                      _SEC_TITLE_APPLIANCE_UPLINKS_USAGE),
                     (_SEC_NAME_APPLIANCE_VPNS, _SEC_TITLE_APPLIANCE_VPNS),
                     (_SEC_NAME_APPLIANCE_PERFORMANCE,
                      _SEC_TITLE_APPLIANCE_PERFORMANCE),
                     (_SEC_NAME_CELLULAR_UPLINKS, _SEC_TITLE_CELLULAR_UPLINKS),
                     (_SEC_NAME_DEVICE_STATUSES, _SEC_TITLE_DEVICE_STATUSES),
                     (_SEC_NAME_DEVICE_UPLINKS_INFO,
                      _SEC_TITLE_DEVICE_UPLINKS_INFO),
                     (_SEC_NAME_LICENSES_OVERVIEW, _SEC_TITLE_LICENSES_OVERVIEW),
                     (_SEC_NAME_SENSOR_READINGS, _SEC_TITLE_SENSOR_READINGS),
                     (_SEC_NAME_SWITCH_PORTS_STATUSES,
                      _SEC_TITLE_SWITCH_PORTS_STATUSES),
                     (_SEC_NAME_WIRELESS_ETHERNET_STATUSES,
                      _SEC_TITLE_WIRELESS_ETHERNET_STATUSES),
                     (_SEC_NAME_WIRELESS_DEVICE_STATUS,
                      _SEC_TITLE_WIRELESS_DEVICE_STATUS),
                     (_SEC_NAME_ORG_SWITCH_PORTS_STATUSES,
                      _SEC_TITLE_ORG_SWITCH_PORTS_STATUSES),
                 ],
                 help=_(
                     'Query only the selected sections. Default is Query all sections.'),
                 default_value=[
                     _SEC_NAME_ORG_SWITCH_PORTS_STATUSES,
                     _SEC_NAME_APPLIANCE_PERFORMANCE,
                     _SEC_NAME_SWITCH_PORTS_STATUSES,
                     _SEC_NAME_WIRELESS_DEVICE_STATUS,
                 ],
             )),
            ('orgs',
             ListOfStrings(
                 title=_('Organisation IDs'),
                 help=_(
                     'List of Organisation IDs to query. Defaulr is all Organisation IDs'),
                 allow_empty=False,
                 validate=_validate_orgs,
             )),
            ('cache_per_section',
             Tuple(
                 title='Set Cache time per section',
                 elements=[
                     Integer(title=_SEC_TITLE_APPLIANCE_PERFORMANCE, minvalue=0, unit='minutes', default_value=0),
                     Integer(title=_SEC_TITLE_APPLIANCE_UPLINKS_USAGE, minvalue=0, unit='minutes', default_value=0),
                     Integer(title=_SEC_TITLE_APPLIANCE_UPLINKS, minvalue=0, unit='minutes', default_value=60),
                     Integer(title=_SEC_TITLE_APPLIANCE_VPNS, minvalue=0, unit='minutes', default_value=60),
                     Integer(title=_SEC_TITLE_CELLULAR_UPLINKS, minvalue=0, unit='minutes', default_value=60),
                     Integer(title=_SEC_TITLE_DEVICE_INFO, minvalue=0, unit='minutes', default_value=60),
                     Integer(title=_SEC_TITLE_DEVICE_STATUSES, minvalue=0, unit='minutes', default_value=60),
                     Integer(title=_SEC_TITLE_DEVICE_UPLINKS_INFO, minvalue=0, unit='minutes', default_value=60),
                     Integer(title=_SEC_TITLE_LICENSES_OVERVIEW, minvalue=0, unit='minutes', default_value=600),
                     Integer(title=_SEC_TITLE_NETWORKS, minvalue=0, unit='minutes', default_value=600),
                     Integer(title=_SEC_TITLE_ORG_API_REQUESTS, minvalue=0, unit='minutes', default_value=0),
                     Integer(title=_SEC_TITLE_ORG_SWITCH_PORTS_STATUSES, minvalue=0, unit='minutes', default_value=0),
                     Integer(title=_SEC_TITLE_ORGANISATIONS, minvalue=0, unit='minutes', default_value=600),
                     Integer(title=_SEC_TITLE_SENSOR_READINGS, minvalue=0, unit='minutes', default_value=0),
                     Integer(title=_SEC_TITLE_SWITCH_PORTS_STATUSES, minvalue=0, unit='minutes', default_value=0),
                     Integer(title=_SEC_TITLE_WIRELESS_DEVICE_STATUS, minvalue=0, unit='minutes', default_value=30),
                     Integer(title=_SEC_TITLE_WIRELESS_ETHERNET_STATUSES, minvalue=0, unit='minutes', default_value=30),
                 ],
             ))
        ],
        optional_keys=True,
        ignored_keys=['sections', 'host_suffix_prefix'],
        required_keys=['excluded_sections', 'api_key'],
    )


rulespec_registry.register(
    HostRulespec(
        group=RulespecGroupDatasourceProgramsApps,
        name='special_agents:cisco_meraki',
        valuespec=_valuespec_special_agent_cisco_meraki,
    )
)
